
#include <string>
#include <fstream>
#include <map>

#include <iostream>
#include <fstream>
#include <sstream>

#include <math.h>

#include "basepairs.hh"
#include "rna_data.hh"

using namespace std;


/* Details for the data structures in BasePairs

basepairs and their probabilities are stored in

  
- arc_probs_
  This is a (sparse) matrix, where we store basepair-probabilities.
  arc_probs_ contains entries for *all* probabilities given in the input

  The following structures are used by the alignment procedure
  for traversing the basepairs. These structures contain only the relevant arcs
  due to the basepair filtering.

- arcs_
  This is a (sparse) matrix, where we can lookup the arc-index
  for a pair of positions. This gives also information whether there is a relevant arc.
- arc_vec_
  The arcs indexed with the same indices as in arcs_
  for each arc, we record the left and right end, the weight, and the "stacking" weight 
- left_
  left_[i] is a list of arcs that have left end i, list only indices
- right_
  right_[j] is a list of arcs that have right end j, list only indices

there is a further structure for stacking probabilites

arc_probs_stack_
this is in analogy to arc_probs_ and contains the probs P[(i,j) /\ (i+1,j-1)]

*/


/*
New strategy for reading basepairs

two stages:

* while reading insert into SparseMatrix
  
* afterwards construct traversal structures

*/


/*! resize data structures that allow fast acces to an arc arcs by
  its left end, its right end, or its left and right end
*/

void BasePairs::resize(uint s) {
    left_.resize(s);
    right_.resize(s);
}

double BasePairs::prob_min() const {
    return params->min_prob;
}

/*
possible simplification:
store in LeftAdjEntry and RightAdjEntry only the index
consequence: sorting has to be rewritten (cannot easily use stl's sort anymore)
*/

/*! compare right ends of arcs (for sorting) */
bool operator < (const LeftAdjEntry &e1,const LeftAdjEntry &e2) {
    return e1.right()<e2.right();
}

/*! compare left ends of arcs (for sorting) */
bool operator < (const RightAdjEntry &e1,const RightAdjEntry &e2) {
    return e1.left()>e2.left();
}

//! sort all adjacency lists in left_ and right_
void BasePairs::sortAdjLists() {

    for (uint i=0; i<left_.size(); ++i)
	sort(left_[i].begin(),left_[i].end());
    for (uint i=0; i<right_.size(); ++i)
	sort(right_[i].begin(),right_[i].end());

}

//! registers an arc (i,j) with probability prob
//! and conditional probability cond_prob
void BasePairs::register_arc(int i, int j) {
    assert(i<j);
    
    int idx=arc_vec_.size();
    
    arc_vec_.push_back(Arc(idx,i,j));
    
    left_[i].push_back(LeftAdjEntry(arc_vec_[idx]));
    right_[j].push_back(RightAdjEntry(arc_vec_[idx]));
    
    arcs_.set(i,j,idx);
}

//! generate the lists arc_vec_, left_, right_ from the matrices arc_probs_
//! and arc_probs_stack_

void BasePairs::generateBPLists(const RnaData &rnadata) {
    
    len_=get_rnadata()->get_sequence().length();

    resize(len_+1);

    // traverse the entries in the prob matrices
    
    // handle all arcs
    for (int i=len_-3; i>=1 ; i--) {
	for (int j=i+3; j<=len_; j++) {
	    
	    double p = rnadata.get_arc_prob(i,j);
	    double cond_p = rnadata.get_arc_stack_prob(i,j);

	    
#define STRONG_FILTER

#ifdef STRONG_FILTER
	    if ( (p >= params->min_prob) || ( exists_arc(i+1,j-1) && (cond_p >= params->min_prob) )) {
		/*
		std::cout <<"("<< i << "," << j<<") "<<p<<" "<<cond_p;
		if ( exists_arc(i+1,j-1) && cond_p >= params->min_prob ) {
		    std::cout <<" stacked";
		    if (!(p >= params->min_prob)) {
			std::cout <<" only";
		    }

		}
		std::cout <<std::endl;
		*/

		register_arc(i,j);
	    }
#else
	    if ( p >= params->min_prob ) {
		register_arc(i,j);
	    } else if ( ( rnadata.get_arc_prob(i+1,j-1) > 0 ) && (cond_p >= params->min_prob) ) {
		register_arc(i,j);
		if (!exists_arc(i+1,j-1)) {
		    //std::cout << "register "<<i<<" "<<j<<" "<<rnadata.get_arc_prob(i+1,j-1) << " " << cond_p << std::endl;
		    register_arc(i+1,j-1);
		}
	    }
#endif
	}
    }
    sortAdjLists();
}


std::ostream &operator <<(std::ostream &out, const Arc &arc) {
    return out <<"("<< arc.idx()<<":"<<arc.left()<<","<<arc.right()<<")"; 
}

