#ifndef RIBOSUM_HH
#define RIBOSUM_HH

#include <string>
#include <fstream>
#include "matrices.hh"
#include "alphabet.hh"

//! read and provide access to ribosum data
//! 
class Ribosum {
public:
    typedef size_t size_type;
    typedef Matrix<double> matrix_t;
private:
    std::string name;
    matrix_t bm;  //!< scores for base matches, 4x4 matrix
    matrix_t am;  //!< scores for basepair/arc matches,
                        //!< 16x16 matrix
protected:

    typedef Alphabet<std::string> alphabet_type;

    std::istream &
    read_matrix(std::istream &in, 
			 matrix_t &mat,
			 const alphabet_type &names) const;
    
    std::ostream &
    write_matrix(std::ostream &out, 
			  const matrix_t &mat,
			  const alphabet_type &alph) const;

    Ribosum() {}
    
    //! reads the standard ribosum file format
    void
    read_ribosum(std::istream &in);
    
private:
    alphabet_type basename_alphabet;
    alphabet_type arcname_alphabet;

    Alphabet<char> char_basename_alphabet;

    //! transform the basename alphabet to alphabet over characters 
    Alphabet<char> make_char_alphabet() const;

public:
    Ribosum(const std::string &filename);
    
    const matrix_t &get_basematch_scores() const {return bm;}
    const matrix_t &get_arcmatch_scores() const {return am;}

    //! get the basename alphabet as alphabet over strings
    const alphabet_type &string_alphabet() const {return basename_alphabet;}
    
    //! get the basename alphabet as alphabet over characters
    const Alphabet<char> &alphabet() const {return char_basename_alphabet;}
    
    const std::string & get_name() const {return name;}
    
    double basematch_score(char i,char j) const {
	return bm(alphabet().idx(i),alphabet().idx(j));;
    }
    
    double arcmatch_score(char i,char j,char k,char l) const {
	return am(alphabet().idx(i)*4+alphabet().idx(j), alphabet().idx(k)*4+alphabet().idx(l));
    }
    
    friend std::ostream & operator << (std::ostream &out, const Ribosum &ribosum);


};

//! extension of the ribosum class that 
//! maintains additional "raw" information
//! in particular frequencies of bases, basematches, basepairs, and arcmatches 
class RibosumFreq : public Ribosum {
public:
    RibosumFreq(const std::string &filename):Ribosum() {
	std::ifstream in(filename.c_str());
	if (!in.good()) {
	    std::cerr << "Cannot open file "<<filename<<" for reading ribosum data."<<std::endl;
	    exit(-1);
	}
	read_ribosum(in);
	
	read_frequencies(in);
	
	in.close();
    }

private:
    
    matrix_t base_probs_;
    matrix_t basepair_probs_;
    matrix_t basematch_probs_;
    matrix_t arcmatch_probs_;

public:
    //! return the probability of a base
    double
    base_prob(char i) const {
	return base_probs_(alphabet().idx(i),0);
    }
    
    //! return the probability of a basepair
    double
    basepair_prob(char i,char j) const {
	return basepair_probs_(alphabet().idx(i),alphabet().idx(j));
    }
    
    //! return the probability of a base match
    double
    basematch_prob(char i,char j) const {
	return basematch_probs_(alphabet().idx(i),alphabet().idx(j));
    }

    //! return the probability of an arcmatch
    double
    arcmatch_prob(char i, char j, char k, char l) const {
	return arcmatch_probs_(alphabet().idx(i)*4+alphabet().idx(j), alphabet().idx(k)*4+alphabet().idx(l));
    }

    void
    read_matrix(std::istream &in, const std::string &header, matrix_t &mat, size_t xdim, size_t ydim);
    
private:
    
    void
    read_frequencies(std::istream &in);
};



#endif //RIBOSUM_HH
